import itertools
import random

import numpy
import pytest

import judy


@pytest.mark.parametrize('shuffled_insertion', [True, False])
def test_insert_query_clear(*, shuffled_insertion: bool) -> None:
    s = judy.JudyIntSet()

    assert len(s) == 0
    assert list(s) == []

    keys = list(range(0, 10_000, 16))

    if shuffled_insertion:
        random.shuffle(keys)

    for key in keys:
        s.add(key)

    for i, k in enumerate(sorted(keys)):
        assert s.by_index(i) == k
        assert s[i] == k

    assert len(s) == 625
    assert list(s) == list(range(0, 10_000, 16))

    for key in range(0, 10_000, 16):
        assert key in s

    for key in range(1, 10_000, 16):
        assert key not in s

    s.clear()
    assert len(s) == 0

    for key in range(0, 10_000, 16):
        assert key not in s


def test_remove() -> None:
    judy_set = judy.JudyIntSet()
    judy_set.discard(10)

    with pytest.raises(KeyError):
        judy_set.remove(10)

    judy_set.add(10)
    assert len(judy_set) == 1

    judy_set.discard(10)
    assert len(judy_set) == 0

    judy_set.add(10)
    assert len(judy_set) == 1

    judy_set.remove(10)
    assert len(judy_set) == 0


def test_iterators() -> None:
    # empty iterator
    judy_set = judy.JudyIntSet()

    with pytest.raises(StopIteration):
        next(iter(judy_set))

    # iterator, starting at empty set,
    # then discovering first item
    it = iter(judy_set)
    judy_set.add(10)
    assert next(it) == 10

    with pytest.raises(StopIteration):
        next(it)

    # iterator finding newly added items when
    # added in order
    it = iter(judy_set)

    for key in range(10, 100, 10):
        judy_set.add(key)
        assert next(it) == key

    with pytest.raises(StopIteration):
        next(it)

    # it will continue to raise StopIteration
    # even if larger keys are added
    judy_set.add(max(judy_set) + 1)

    with pytest.raises(StopIteration):
        next(it)

    # this won't work for descending keys
    judy_set.clear()
    it = iter(judy_set)

    judy_set.add(1000)
    assert next(it) == 1000

    judy_set.add(2000)
    assert next(it) == 2000

    judy_set.add(0)

    with pytest.raises(StopIteration):
        next(it)


def test_overwrite() -> None:
    judy_set = judy.JudyIntSet()
    judy_set.add(0)
    judy_set.add(0)
    assert list(judy_set) == [0]


def test_str_repr() -> None:
    judy_set = judy.JudyIntSet()

    for _ in range(1000):
        judy_set.add(random.randint(0, 2**64 - 1))

    assert str(judy_set) == f'JudyIntSet({sorted(judy_set)})'
    assert repr(judy_set) == f'JudyIntSet({sorted(judy_set)})'


def test_ranged_iterators() -> None:
    judy_set = judy.JudyIntSet()
    py_set = {random.randint(0, 10_000) for _ in range(1000)}

    for key in py_set:
        judy_set.add(key)

    assert len(judy_set) == len(py_set)
    assert list(judy_set) == sorted(py_set)

    lowers = list(range(0, 10_000, 1000))
    uppers = list(range(0, 10_000, 1000))

    for lower in lowers:
        assert sorted(key for key in py_set if key >= lower) == list(judy_set.ranged_iterator(lower_inclusive=lower))

    for upper in uppers:
        assert sorted(key for key in py_set if key <= upper) == list(judy_set.ranged_iterator(upper_inclusive=upper))

    for lower, upper in itertools.product(lowers, uppers):
        assert sorted(key for key in py_set if lower <= key <= upper) == list(
            judy_set.ranged_iterator(lower_inclusive=lower, upper_inclusive=upper)
        )


def test_pop() -> None:
    judy_set = judy.JudyIntSet()
    py_set = {random.randint(0, 10_000) for _ in range(100)}
    popped = set()

    for key in py_set:
        judy_set.add(key)

    for _ in range(len(judy_set)):
        popped.add(judy_set.pop())

    with pytest.raises(KeyError):
        judy_set.pop()

    assert popped == py_set
    assert len(judy_set) == 0


def test_from_array() -> None:
    n = 1000

    a8 = numpy.random.randint(low=0, high=2**8 - 1, size=n, dtype=numpy.uint8)
    assert sorted(set(a8.tolist())) == sorted(judy.JudyIntSet.FromArray(a8))

    a16 = numpy.random.randint(low=0, high=2**16 - 1, size=n, dtype=numpy.uint16)
    assert sorted(set(a16.tolist())) == sorted(judy.JudyIntSet.FromArray(a16))

    a32 = numpy.random.randint(low=0, high=2**32, size=n, dtype=numpy.uint32)
    assert sorted(set(a32.tolist())) == sorted(judy.JudyIntSet.FromArray(a32))

    a64 = numpy.random.randint(low=0, high=2**64 - 1, size=n, dtype=numpy.uint64)
    assert sorted(set(a64.tolist())) == sorted(judy.JudyIntSet.FromArray(a64))


@pytest.mark.skip(reason='This test is currently under development')
def test_concurrent_mutations() -> None:
    raise NotImplementedError


def test_sizeof() -> None:
    s = judy.JudyIntSet()
    empty = s.__sizeof__()

    for _ in range(1000):
        s.add(_)

    bigger = s.__sizeof__()
    assert empty < bigger



import itertools
import random
from typing import Any, TypeAlias

import pytest

import judy

from .test_utils import judy_dict_contains_items, key_value_gen


JudyMap: TypeAlias = judy.JudyIntIntMap | judy.JudyIntObjectMap[Any]
JudyMaps: tuple[type[JudyMap], ...] = (judy.JudyIntIntMap, judy.JudyIntObjectMap)


@pytest.mark.parametrize('klass', JudyMaps)
def test_insert_query_clear(klass: type[JudyMap]) -> None:
    # basic insert, query, clear
    m = klass()

    key_range = range(0, 16_000, 16)
    value_range = range(16_000, 32_000, 16)

    for key, value in key_value_gen(keys=key_range, values=value_range, shuffle=True):
        with pytest.raises(IndexError):
            m.by_index(len(m))

        m[key] = value

    assert len(m) == len(key_range)
    assert list(m.keys()) == list(key_range)

    for i, (key, value) in enumerate(zip(iter(key_range), iter(value_range))):
        assert key in m
        assert m[key] == value
        assert m.get(key) == value
        assert m.get(key, 1337) == value
        assert m.by_index(i) == value

    judy_dict_contains_items(m, list(zip(iter(key_range), iter(value_range))))

    for key in range(1, 16_000, 16):
        assert key not in m

        with pytest.raises(KeyError):
            assert m[key]

        with pytest.raises(KeyError):
            assert m.pop(key)

        assert m.pop(key, None) is None

    m.clear()
    assert len(m) == 0

    for key in iter(key_range):
        assert key not in m

        with pytest.raises(KeyError):
            assert m[key] == value

        assert m.get(key) is None
        assert m.get(key, 1337) == 1337


@pytest.mark.parametrize('klass', JudyMaps)
def test_ranged_iterators(klass: type[JudyMap]) -> None:
    judy_map = klass()
    UPPER = 10_000
    COUNT = 10_000
    RANGE_INC = 1_000

    keys = {random.randint(0, UPPER) for _ in range(COUNT)}

    for key in keys:
        judy_map[key] = key * 2

    s_keys = sorted(keys)
    s_values = [key * 2 for key in s_keys]

    assert len(judy_map) == len(keys)
    assert list(judy_map) == sorted(keys)

    lowers = list(range(0, UPPER, RANGE_INC))
    uppers = list(range(0, UPPER, RANGE_INC))

    def _in_range(*, lower_inclusive: int | None = None, upper_inclusive: int | None = None) -> list[tuple[int, int]]:
        return [
            (key, value)
            for key, value in zip(s_keys, s_values)
            if (lower_inclusive is None or key >= lower_inclusive)
            and (upper_inclusive is None or key <= upper_inclusive)
        ]

    for lower in lowers:
        assert [kv[0] for kv in _in_range(lower_inclusive=lower)] == list(judy_map.keys(lower_inclusive=lower))
        assert [kv[1] for kv in _in_range(lower_inclusive=lower)] == list(judy_map.values(lower_inclusive=lower))
        assert _in_range(lower_inclusive=lower) == list(judy_map.items(lower_inclusive=lower))

    for upper in uppers:
        assert [kv[0] for kv in _in_range(upper_inclusive=upper)] == list(judy_map.keys(upper_inclusive=upper))
        assert [kv[1] for kv in _in_range(upper_inclusive=upper)] == list(judy_map.values(upper_inclusive=upper))
        assert _in_range(upper_inclusive=upper) == list(judy_map.items(upper_inclusive=upper))

    for lower, upper in itertools.product(lowers, uppers):
        assert [kv[0] for kv in _in_range(lower_inclusive=lower, upper_inclusive=upper)] == list(
            judy_map.keys(lower_inclusive=lower, upper_inclusive=upper)
        )
        assert [kv[1] for kv in _in_range(lower_inclusive=lower, upper_inclusive=upper)] == list(
            judy_map.values(lower_inclusive=lower, upper_inclusive=upper)
        )
        assert _in_range(lower_inclusive=lower, upper_inclusive=upper) == list(
            judy_map.items(lower_inclusive=lower, upper_inclusive=upper)
        )


@pytest.mark.parametrize('klass', JudyMaps)
def test_set_del(klass: type[JudyMap]) -> None:
    m = klass()
    assert len(m) == 0

    m[10] = 20
    assert len(m) == 1

    m[10] = 30
    assert len(m) == 1


@pytest.mark.parametrize('klass', JudyMaps)
def test_sizeof(klass: type[JudyMap]) -> None:
    s = klass()
    empty = s.__sizeof__()

    for _ in range(1000):
        s[_] = _ * 10_000

    bigger = s.__sizeof__()
    assert empty < bigger



import sys

import judy


def test_refleaks() -> None:
    obj = object()

    def ref() -> int:
        return sys.getrefcount(obj) - 1

    expected = ref()

    m = judy.JudyIntObjectMap()
    n = 1000

    for _ in range(n):
        m[_] = obj
        expected += 1
        assert expected == ref()

    for _ in range(n // 2):
        del m[_]
        expected -= 1
        assert expected == ref()

    # pop
    m.pop(700)
    expected -= 1
    assert expected == ref()

    # add
    m[700] = obj
    expected += 1
    assert expected == ref()

    # overwrite
    m[700] = obj
    assert expected == ref()

    # failed-pop with default
    m.pop(20000, None)
    assert expected == ref()

    i = m.by_index(10)
    expected += 1
    assert expected == ref()

    del i
    expected -= 1
    assert expected == ref()

    pre_iter = ref()

    for it in (m.values(), m.items()):
        assert expected == ref()

        for _ in it:
            expected += 1
            assert expected == ref()
            expected -= 1

        del _
        assert pre_iter == ref()
        assert expected == ref()

    expected -= len(m)
    m.clear()
    assert expected == ref()


def test_circular_destruction() -> None:
    m = judy.JudyIntObjectMap()

    def ref() -> int:
        return sys.getrefcount(m) - 1  # noqa: F821

    expected = ref()

    m[1] = m
    expected += 1
    assert expected == ref()

    m.clear()
    expected -= 1

    assert expected == ref()

    m[1] = m
    del m


def test_str_repr() -> None:
    m = judy.JudyIntObjectMap()
    assert str(m) == 'JudyIntObjectMap({})'
    assert repr(m) == 'JudyIntObjectMap({})'

    m[1] = m
    assert str(m) == 'JudyIntObjectMap({1: JudyIntObjectMap({...})})'
    assert repr(m) == 'JudyIntObjectMap({1: JudyIntObjectMap({...})})'

    k = judy.JudyIntObjectMap()
    k[1337] = m
    assert str(k) == 'JudyIntObjectMap({1337: JudyIntObjectMap({1: JudyIntObjectMap({...})})})'
    assert repr(k) == 'JudyIntObjectMap({1337: JudyIntObjectMap({1: JudyIntObjectMap({...})})})'

    m.clear()
    m[1] = 'hello python 🐍'
    assert str(m) == 'JudyIntObjectMap({1: \'hello python 🐍\'})'
    assert repr(m) == 'JudyIntObjectMap({1: \'hello python 🐍\'})'



import random
from collections.abc import Sequence

import judy


def key_value_gen(keys: range, values: range, *, shuffle: bool) -> list[tuple[int, int]]:
    assert len(keys) == len(values)
    keys_and_values = list(zip(iter(keys), iter(values)))

    if shuffle:
        random.shuffle(keys_and_values)

    return keys_and_values


def judy_dict_contains_items(
    m: judy.JudyIntIntMap | judy.JudyIntObjectMap[int], sorted_dedup_items: Sequence[tuple[int, int]]
) -> None:
    assert len(m) == len(sorted_dedup_items)
    assert list(m) == [k for k, _ in sorted_dedup_items]
    assert list(m.keys()) == [k for k, _ in sorted_dedup_items]
    assert list(m.items()) == list(sorted_dedup_items)
    assert list(m.values()) == [v for _, v in sorted_dedup_items]


def judy_set_contains_items(m: judy.JudyIntSet, sorted_dedup_items: Sequence[tuple[int, int]]) -> None:
    assert len(m) == len(sorted_dedup_items)
    assert list(m) == [k for k, _ in sorted_dedup_items]
    assert list(m.keys()) == [k for k, _ in sorted_dedup_items]
    assert list(m.items()) == list(sorted_dedup_items)
    assert list(m.values()) == [v for _, v in sorted_dedup_items]



from typing import TypeAlias

import pytest

import judy


JudyContainer: TypeAlias = judy.JudyIntSet | judy.JudyIntIntMap | judy.JudyIntObjectMap
JudyContainers: tuple[type[JudyContainer], ...] = (judy.JudyIntSet, judy.JudyIntIntMap, judy.JudyIntObjectMap)


@pytest.mark.parametrize('klass', JudyContainers)
def test_empty_by_index(klass: type[JudyContainer]) -> None:
    j = klass()

    for index in (-100, 100):
        with pytest.raises(IndexError):
            j.by_index(index)


@pytest.mark.parametrize('klass', JudyContainers)
def test_empty_clear(klass: type[JudyContainer]) -> None:
    j = klass()
    assert len(j) == 0
    assert list(j) == []
    j.clear()
    assert len(j) == 0


@pytest.mark.parametrize('klass', JudyContainers)
def test_compare(klass: type[JudyContainer]) -> None:
    s = klass()
    t = klass()

    with pytest.raises(TypeError):
        s < t  # noqa: B015

    with pytest.raises(TypeError):
        s > t  # noqa: B015

    with pytest.raises(TypeError):
        s <= t  # noqa: B015

    with pytest.raises(TypeError):
        s >= t  # noqa: B015

    with pytest.raises(TypeError):
        s == t  # noqa: B015

    with pytest.raises(TypeError):
        s != t  # noqa: B015



import numpy
import pytest

import judy

from .test_utils import judy_set_contains_items


def test_empty() -> None:
    s = judy.JudyIntIntMap()
    assert len(s) == 0

    for _ in range(2):
        assert [] == list(s)
        assert [] == list(s.keys())
        assert [] == list(s.values())
        assert [] == list(s.items())
        s.clear()

    judy_set_contains_items(s, [])


def test_from_array() -> None:
    n = 1000

    a64_empty = numpy.random.randint(low=0, high=2**64 - 1, size=0, dtype=numpy.uint64)

    # we only support 64-bit variants
    keys = numpy.random.randint(low=0, high=2**64 - 1, size=n, dtype=numpy.uint64)
    values = numpy.random.randint(low=0, high=2**64 - 1, size=n, dtype=numpy.uint64)

    py_m = {k.item(): v.item() for k, v in zip(keys, values)}
    py_k = sorted(py_m)
    py_v = [v for k, v in sorted(py_m.items())]
    py_kv = list(zip(py_k, py_v))

    m = judy.JudyIntIntMap.FromArray(keys, values)

    assert py_k == list(m)
    assert py_k == list(m.keys())
    assert py_v == list(m.values())
    assert py_kv == list(m.items())

    with pytest.raises(RuntimeError):
        judy.JudyIntIntMap.FromArray(a64_empty, values)


def test_str_repr() -> None:
    m = judy.JudyIntIntMap()

    for _ in range(10):
        m[_] = _ * 2

    payload = '{' + ', '.join(f'{k}: {v}' for k, v in m.items()) + '}'

    assert str(m) == f'JudyIntIntMap({payload})'
    assert repr(m) == f'JudyIntIntMap({payload})'
